/**
@file 
@date        12/07/2002
@author      Herve Milcent
@copyright   CERN (AB-CO)

@brief Library of the Laser component used by the systemIntegrity.
	 	
Usage: Public

PVSS manager usage: CTRL, UI

Constraints:
- global variable: the following variables are global to the script
- data point type needed: _UnSystemIntegrity
- data point: Laser_systemIntegrity
- PVSS version: 2.12.1 
- operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
- distributed system: yes.

@TODO: Simplify this plugin, because information flow in current implementation is way too complex to be easily maintainable
@TODO: Think about synchronisation mechanisms
*/

// constant declaration
private const string UN_SYSTEM_INTEGRITY_Laser = "Laser";
public const int c_unSystemIntegrity_Laser_alarmValue_T_mode = 12;  
public const int c_unSystemIntegrity_Laser_alarmValue_Unknown = 1300;  

// global declaration
private global dyn_string g_dsLaser_SystemAlarmList;    // List of active LASER System-Alarm-DPs
private global dyn_int g_diLaser_ActiveApiManagers;		   // List of active LASER Api-Managers

// These globals are also used by the unLaserStatusList.pnl
public global mapping g_mLaser_MessageOut;      // Output message of the LASER Manager
public global mapping g_mLaser_CommandIn;       // Input command of the LASER Manager
public global mapping g_mLaser_StateOut;        // Operating state of the LASER Manager: running, active, passive, stopped, ...
public global mapping g_mLaser_CheckPeriod;     // System integrity check period of the LASER Manager
public global mapping g_mLaser_Tmode;           // T Mode state of the LASER Manager
public global mapping g_mLaser_AlertValue;      // Alert value of the LASER Manager
public global mapping g_mLaser_StatePvss;       // State from PVSS-Console of the LASER Manager
public global mapping g_mLaser_Thread;          // Monitoring thread of the LASER Manager        


/** 
* Returns the list of systemAlarm patterns. 
*
* @return list of systemAlarm patterns
*/
public dyn_string Laser_systemIntegrityInfo()
{
    return makeDynString(LASER_pattern);
}

/** 
* Get list of laser managers checked by the systemIntegrity. The ones that are enabled.
* 
* @param dsResult: dyn_string, output, the list of enabled Laser that are checked
*/
public void unSystemIntegrity_Laser_Initialize(dyn_string &dsResult)
{
    string sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("unLHCLASER"));
		
    // get the list of Laser manager dp to checked
    dyn_string dsList = dpNames(c_unSystemAlarm_dpPattern+LASER_pattern+"*", c_unSystemAlarm_dpType);
    int len = dynlen(dsList);
    for(int i = 1; i<=len; i++) {        
		      string dpToCheck = dpSubStr(dsList[i], DPSUB_DP);
		      dpToCheck = substr(dpToCheck, strpos(dpToCheck,LASER_pattern));        
        bool enabled;
		      dpGet(dsList[i]+".enabled", enabled);
        if(enabled) {
			         dynAppend(dsResult, dpToCheck);
        }
    }

    dyn_string dsExInfo;
    unMessageText_send("*", "*", g_SystemIntegrity_sSystemName+"systemIntegrity", "user", "*", "INFO", "systemIntegrity "+sMessage+" loaded", dsExInfo);
}

/** 
* Handle systemIntegrity commands.
* 
* @param sDpe1: string, input, dpe
* @param command: int, input, the systemIntegrity command
* @param sDpe2: string, input, dpe
* @param parameters: dyn_string, input, the list of parameters of the systemIntegrity command
*/
public void unSystemIntegrity_Laser_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters)
{
    dyn_string exceptionInfo;
    int i, len=dynlen(parameters);
    string sMessage;
	
    switch(command) 
    {
		  case UN_SYSTEMINTEGRITY_ADD: // add the _UnSystemAlarm DP
			     for(i=1; i<=len; i++) {
				        unSystemIntegrity_Laser_checking(parameters[i], true, true, exceptionInfo);
			     }
			     break;
		  case UN_SYSTEMINTEGRITY_DELETE: // delete the _UnSystemAlarm DP
			     for(i=1; i<=len; i++) {
    				    unSystemIntegrity_Laser_checking(parameters[i], false, false, exceptionInfo);
    				    // remove dp from the alert list of applicationDP if it is in
    			    	_unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern+parameters[i], false, exceptionInfo);
    			    	// delete the _UnSystemAlarm dps
    				    if(dpExists(c_unSystemAlarm_dpPattern+parameters[i]))
    					       dpDelete(c_unSystemAlarm_dpPattern+parameters[i]);
			     }
			     break;
		  case UN_SYSTEMINTEGRITY_ENABLE: // enable the _UnSystemAlarm DP
    			 for(i=1; i<=len; i++) {
    				if(dpExists(c_unSystemAlarm_dpPattern+parameters[i]))
    				 	  unSystemIntegrity_Laser_checking(parameters[i], false, true, exceptionInfo);
    			 }
    		 	break;
  		case UN_SYSTEMINTEGRITY_DISABLE: // disable the _UnSystemAlarm DP
			     for(i=1; i<=len; i++) {
				        if(dpExists(c_unSystemAlarm_dpPattern+parameters[i]))
					           unSystemIntegrity_Laser_checking(parameters[i], false, false, exceptionInfo);
			     }
			     break;
  		case UN_SYSTEMINTEGRITY_DIAGNOSTIC: // give the list of _UnSystemAlarm DP and the state
			     dpSet(UN_SYSTEM_INTEGRITY_Laser+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
					         UN_SYSTEM_INTEGRITY_Laser+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", g_dsLaser_SystemAlarmList);
			     break;
    case UN_SYSTEMINTEGRITY_VERSION:
        sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("unLHCLASER"));
        dpSet(UN_SYSTEM_INTEGRITY_Laser+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
              UN_SYSTEM_INTEGRITY_Laser+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", makeDynString(getCurrentTime(),sMessage));
        break;
  		default:
			     break;
    }

    if(dynlen(exceptionInfo) > 0) {
        if(isFunctionDefined("unMessageText_sendException")) {
			         unMessageText_sendException("*", "*", __FUNCTION__, "user", "*", exceptionInfo);
		      }
        // handle error in case send message failed
		      if(dynlen(exceptionInfo) > 0) {
			         DebugTN(getCurrentTime(), exceptionInfo);
		      }
    }
}

/**
* Registers and de-registers monitoring threads and callback functions for a specific laser manager. 
*
* It can also create the _unSystemAlarm_Laser dp with the alarm config but it cannot delete it.
*
* @param sDp_Num: string, input, data point name
* @param bCreate: bool, input, true create the dp and the alarm config if it does not exist, enable it
* @param bRegister: bool, input, true do a dpConnect, false do a dpDisconnect
* @param dsExInfo: dyn_string, output, exception are returned here
*/
private void unSystemIntegrity_Laser_checking(string sDp_Num, bool bCreate, bool bRegister, dyn_string &dsExInfo)
{	
    string dp_Alarm = c_unSystemAlarm_dpPattern + sDp_Num;
	
	   sDp_Num=substr(sDp_Num,strpos(sDp_Num,"_")+1);    
    string sNum = (strlen(sDp_Num)<2 ? "0" : "") + sDp_Num;	
	   string sManager = LASER_pattern + "Manager_" + sNum;
    
    string sSystem = getSystemName();
   	string dpeMessageOut = sSystem + sManager + ".State.ev_MessageOut";
 	  string dpeTmode = sSystem + sManager + ".State.T_mode";
    
	   if (bRegister)
	   {
        // no need to register, if already existent
        if(dynContains(g_dsLaser_SystemAlarmList, dp_Alarm) > 0 || !dpExists(sSystem+sManager)) 
            return;
      
        int iRes;
      		if (dynlen(g_dsLaser_SystemAlarmList)==0) //only ONE Api-Listener for all Managers
		          iRes = unSystemIntegrity_Laser_registerApiManagerListener();
			
      		if (iRes>=0)
      		{
            unSystemIntegrity_Laser_addManagerToGlobalMapping(sManager);
            unSystemIntegrity_Laser_registerSettingsListener(sManager); 		
				
        			 //Create the dp_Alarm if it is not existing
        			 if(!dpExists(dp_Alarm)) {
        			     string description = getCatStr("unSystemIntegrity", "LASER_DESCRIPTION") + sDp_Num;
        			   	 unSystemIntegrity_createSystemAlarm(sDp_Num, LASER_pattern, description, dsExInfo);
        			 }
				
      		 	  //Connect
      			   if(dynlen(dsExInfo)<=0) {                    
                unSystemIntegrity_Laser_addMonitoringThread(sManager); //Start Thread to check Period
						
				            _unSystemIntegrity_registerFunction(
                        bRegister, 
                        "unSystemIntegrity_Laser_MonitoringCB", 
                        makeDynString(dpeMessageOut, dpeTmode), 
						                  g_dsLaser_SystemAlarmList, 
                        dp_Alarm, 
                        dsExInfo);
            }
        }		
    }
    else // unregister
    {
        // no need to delete, if not registered 
        if(dynContains(g_dsLaser_SystemAlarmList, dp_Alarm) <= 0) 
            return;        
        
        unSystemIntegrity_Laser_removeMonitoringThread(sManager);
        unSystemIntegrity_Laser_unregisterSettingsListener(sManager);
			
		      _unSystemIntegrity_registerFunction(
                bRegister, 
                "unSystemIntegrity_Laser_MonitoringCB", 
                makeDynString(dpeMessageOut, dpeTmode), 
                g_dsLaser_SystemAlarmList, 
                dp_Alarm, 
                dsExInfo);
			
        if(dynlen(dsExInfo)<=0) {
            unSystemIntegrity_Laser_removeManagerFromGlobalMapping(sManager);		
		      }
			
        if (dynlen(g_dsLaser_SystemAlarmList)==0) {
		          unSystemIntegrity_Laser_unregisterApiManagerListener();
        }
    }
} 

/**
* Registers callback, that updates gloabal heart-beat-period of the laser manager
*
* @param sManager  input, name of the laser manager
* @param sSystem   input, system name
*/
public int unSystemIntegrity_Laser_registerSettingsListener(string sManager, string sSystem=getSystemName()) {
    return dpConnect("unSystemIntegrity_Laser_SettingsListenerCB", sSystem+sManager+".Settings.man_NumPeriod");       
}

/**
* Unregisters callback, that updates gloabal heart-beat-period of the laser manager
*
* @param sManager  input, name of the laser manager
* @param sSystem   input, system name
*/
public int unSystemIntegrity_Laser_unregisterSettingsListener(string sManager, string sSystem=getSystemName()) {
    return  dpDisconnect("unSystemIntegrity_Laser_SettingsListenerCB", sSystem+sManager+".Settings.man_NumPeriod");     
}

/**
* Callback that updates gloabal heart-beat-period of the laser manager
*
* @param sDpeHeartBeatPeriod  input, dpe storing the heart-beat-period
* @param iHeartBeatPeriod     input, value of the heart-beat-period
*/
public void unSystemIntegrity_Laser_SettingsListenerCB(string sDpeHeartBeatPeriod, int iHeartBeatPeriod) {   
    string sManager = dpSubStr(sDpeHeartBeatPeriod, DPSUB_DP);	    
    g_mLaser_CheckPeriod[sManager] = iHeartBeatPeriod * 2;   
    unSystemIntegrity_Laser_ensureValidCheckPeriod(sManager);   
    DebugFTN(1, __FUNCTION__, g_mLaser_CheckPeriod);
}

/**
* Registers callback that checks which laser manager numbers are active right now
*
* @param sSystem   input, system name
*/
public int unSystemIntegrity_Laser_registerApiManagerListener(string sSystem=getSystemName()) {
    return dpConnect("unSystemIntegrity_Laser_ApiManagerListenerCB", sSystem+"_Connections.Api.ManNums");   
}

/**
* Unregisters callback that checks which laser manager numbers are active right now
*
* @param sSystem   input, system name
*/
public int unSystemIntegrity_Laser_unregisterApiManagerListener(string sSystem=getSystemName()) {
    return dpDisconnect("unSystemIntegrity_Laser_ApiManagerListenerCB", sSystem+"_Connections.Api.ManNums");   
}

/**
* Callback that checks which laser manager numbers are active right now
*
* @param sDp_1                   input, dpe, not used
* @param diApiManagerNumbers     input, array of manager numbers active right now
*/
public void unSystemIntegrity_Laser_ApiManagerListenerCB(string sDp_1, dyn_int diApiManagerNumbers) {
    g_diLaser_ActiveApiManagers = diApiManagerNumbers;
    DebugFTN(1, __FUNCTION__, g_diLaser_ActiveApiManagers);
}

/**
* Checks whether a specific laser manager is active right now 
*
* @param sManager                   input, name of the laser manager
*
* @return true: manager is active, false: manager is not active
*/
public bool unSystemIntegrity_Laser_isApiManagerActive(string sManager) {
    int iManagerNumber = unLaser_getManagerNumberFromName(sManager);        
    if(dynContains(g_diLaser_ActiveApiManagers, iManagerNumber) > 0) {
        return true;
    }     
    return false;               
}

/**
* Adds a monitoring thread for a specific laser manager
*
* @param sManager                   input, name of the laser manager
*/
private void unSystemIntegrity_Laser_addMonitoringThread(string sManager) {
    //Start Thread to check Period
    int threadId = startThread("unSystemIntegrity_Laser_MonitoringThread", sManager);
    if (threadId >= 0) {
        g_mLaser_Thread[sManager] = threadId; 
    }    
}

/**
* Removes a monitoring thread for a specific laser manager
*
* @param sManager                   input, name of the laser manager
*/
private void unSystemIntegrity_Laser_removeMonitoringThread(string sManager) {      
    if(mappingHasKey(g_mLaser_Thread, sManager)) {
        int threadId = g_mLaser_Thread[sManager];  
        int iRes=stopThread(threadId); 
        if (iRes==0) {
            mappingRemove(g_mLaser_Thread, sManager);   
        }
    }   
}

/**
* Add a new laser manager to global mappings
*
* These global mappings store informations about running laser managers
*
* @param sManager                   input, name of the laser manager
*/
public void unSystemIntegrity_Laser_addManagerToGlobalMapping(string sManager) {    
    g_mLaser_MessageOut[sManager] = 0;
    g_mLaser_CommandIn[sManager] = 0;
    g_mLaser_StateOut[sManager] = 0;   
    g_mLaser_CheckPeriod[sManager] = 0;    
    g_mLaser_Tmode[sManager] = 0;
    g_mLaser_AlertValue[sManager] = -1;    
    g_mLaser_StatePvss[sManager] = 0;
}

/**
* Remove a laser manager from global mappings
*
* These global mappings store informations about running laser managers
*
* @param sManager                   input, name of the laser manager
*/
private void unSystemIntegrity_Laser_removeManagerFromGlobalMapping(string sManager) {  
    if(mappingHasKey(g_mLaser_MessageOut, sManager)) mappingRemove(g_mLaser_MessageOut, sManager);    
    if(mappingHasKey(g_mLaser_CommandIn, sManager)) mappingRemove(g_mLaser_CommandIn, sManager);      
    if(mappingHasKey(g_mLaser_StateOut, sManager)) mappingRemove(g_mLaser_StateOut, sManager);   
    if(mappingHasKey(g_mLaser_CheckPeriod, sManager)) mappingRemove(g_mLaser_CheckPeriod, sManager);         
    if(mappingHasKey(g_mLaser_Tmode, sManager)) mappingRemove(g_mLaser_Tmode, sManager);     
    if(mappingHasKey(g_mLaser_AlertValue, sManager)) mappingRemove(g_mLaser_AlertValue, sManager);          
    if(mappingHasKey(g_mLaser_StatePvss, sManager)) mappingRemove(g_mLaser_StatePvss, sManager);   
}

/**
* Remove all laser managers from global mapping
*
* These global mappings store informations about running laser managers
*/
public void unSystemIntegrity_Laser_clearGlobalMapping() {
    mappingClear(g_mLaser_MessageOut);
    mappingClear(g_mLaser_CommandIn);
    mappingClear(g_mLaser_StateOut);   
    mappingClear(g_mLaser_CheckPeriod);    
    mappingClear(g_mLaser_Tmode);
    mappingClear(g_mLaser_AlertValue);    
    mappingClear(g_mLaser_StatePvss);    
}

/**
* Thread, constantly checks health of a laser manager in a specific interval
*
* @param sManager                   input, name of the laser manager
* @param sSystemName                input, name of the system
*/
private void unSystemIntegrity_Laser_MonitoringThread(string sManager, string sSystemName=getSystemName()) { 
    
    unSystemIntegrity_Laser_ensureValidCheckPeriod(sManager);	
    
    while (true) 	
    {
        bool stateChanged = unSystemIntegrity_Laser_updateGlobalManagerState(sManager, sSystemName);
        
        if (stateChanged) {  
            // Trigger unSystemIntegrity_Laser_MonitoringCB
            dpSet(sSystemName + sManager + ".State.ev_MessageOut", g_mLaser_MessageOut[sManager]); 	           
        } 
        
        delay( unSystemIntegrity_Laser_getCheckPeriod(sManager) ); 
    } 
}

/**
* Ensures that check interval of the monitoring thread is at least min. rate
*
* @param sManager                   input, name of the laser manager
*/
public void unSystemIntegrity_Laser_ensureValidCheckPeriod(string sManager) { 
    if (g_mLaser_CheckPeriod[sManager] < c_Min_Rate) {
		      g_mLaser_CheckPeriod[sManager] = c_Min_Rate;  
    }  
}

/**
* Gets the current check interval of the monitoring thread of a laser manager
*
* @param sManager                   input, name of the laser manager
*/
public unsigned unSystemIntegrity_Laser_getCheckPeriod(string sManager) { 
    if (mappingHasKey(g_mLaser_CheckPeriod, sManager)) {
        return g_mLaser_CheckPeriod[sManager];
    }    
    return c_Min_Rate;    
}

/**
* Updates the current state of a laser mangager and returns true if the state changed
*
* @param sManager                   input, name of the laser manager
* @param sSystemName                input, name of the system
*
* @return true: state changed, false: state stayed the same
*/
public bool unSystemIntegrity_Laser_updateGlobalManagerState(string sManager, string sSystemName)
{
    int statePvss = unSystemIntegrity_Laser_getManagerStateConsole(sManager, sSystemName);
        
    if (g_mLaser_StatePvss[sManager] != statePvss) { // state has changed
        g_mLaser_StatePvss[sManager] = statePvss;	
        return true;
    }
    return false;    
}

/**
* Retrieves the current state of a laser mangager
*
* @param sManager                   input, name of the laser manager
* @param sSystemName                input, name of the system
*
* @return state: 0=stopped, 1=running, 2=blocked, -1=error
*/
private int unSystemIntegrity_Laser_getManagerStateConsole(string sManager, string sSystemName)
{    	    
    if(!unSystemIntegrity_Laser_isApiManagerActive(sManager))
        return 0; // Stopped
                 
    // Take heart-beat of the laser manager into account
    
    time tLastRefresh;      
    string sDpeManNum = sSystemName + sManager + ".Settings.man_Num";  
    int iRes = dpGet(sDpeManNum + ":_online.._stime", tLastRefresh);
	int iLastPeriod = period(tLastRefresh); // seconds from epoch to last refresh
	int iNowPeriod = period(getCurrentTime()); // seconds from epoch to now
    if (iRes < 0 || iLastPeriod < 0 || iNowPeriod < 0)
        return -1; // Error   
    
    int iPeriod = iNowPeriod-iLastPeriod;
    if (iPeriod > g_mLaser_CheckPeriod[sManager])  
       return 2; // blocked
    
    return 1; // running
}

/** 
* Callback to monitor laser manager health, triggers sytem alarm on error
*
* In this callback function all available health information about a laser manager (stored in global mappings)
* is being put together and evaluated and a system alarm is being triggered on error
*
*	@param sDpeMessageOut: input, string, data point name
*	@param iMessageOut:    input, int, Message from Laser mamager
*	@param sDpeTmode:      input, string, data point name
*	@param Tmode:          input, int, Test mode state 
*
*/
private void unSystemIntegrity_Laser_MonitoringCB(string sDpeMessageOut, int iMessageOut, string sDpeTmode, bool Tmode)
{
    string sManager = substr(sDpeMessageOut, strpos(sDpeMessageOut, c_Laser_Pattern), strlen(c_Laser_Pattern) + 2);   
	         
    g_mLaser_MessageOut[sManager] = iMessageOut;    
    g_mLaser_Tmode[sManager] = Tmode;
    
    int alertValue = unSystemIntegrity_Laser_getAlertValue(sManager); 
    
    // set alarm, if alert value changes
    if (g_mLaser_AlertValue[sManager] != alertValue) { 
        g_mLaser_AlertValue[sManager] = alertValue;
		        
        unSystemIntegrity_Laser_setAlarm(sManager, alertValue);
    }	
    
    DebugFTN(1, __FUNCTION__, g_mLaser_MessageOut, g_mLaser_Tmode, g_mLaser_StatePvss, g_mLaser_AlertValue);		
}

/** 
* Calculates alert value for a specific laser manager using data in global mappings
*
*	@param sDpeMessageOut: input, name of the laser manager
* @return alert value
*/
public int unSystemIntegrity_Laser_getAlertValue(string sManager)
{
    // Take manager state from PVSS console into account
    switch(g_mLaser_StatePvss[sManager])
    {
    case 0:	// Manager stopped
	       return c_Stopped;	
                						
	   case 1:	// Manager Started                     
        break;	
                			
	   case 2:	// Manager blocked
	       return c_Blocked;		
                		
	   default: // Unknown error	
	       return c_unSystemIntegrity_Laser_alarmValue_Unknown;		
    }   
     
    // Take manager output message into account if it is at least a warning
    if (g_mLaser_MessageOut[sManager] >= LASER_MANAGER_MSG_WARNING_BASE) {	
        return g_mLaser_MessageOut[sManager];
    } 
    
    // Take Tmode into account
    if(g_mLaser_Tmode[sManager]) {
        return c_unSystemIntegrity_Laser_alarmValue_T_mode;    
    }
    
    return c_unSystemIntegrity_no_alarm_value;  
}

/** 
* Sets the system alarm datapoint for a specific laser manager to a specific alertValue
*
*	@param sDpeMessageOut: input, name of the laser manager
*	@param alertValue:     input, alert value
*/
private void unSystemIntegrity_Laser_setAlarm(string sManager, int alertValue) 
{
    int iManagerNumber = unLaser_getManagerNumberFromName(sManager);
    string sAlarmDpe = c_unSystemAlarm_dpPattern + LASER_pattern + iManagerNumber + ".alarm"; 
        				
    if (dpExists(sAlarmDpe)) {
	       dpSetWait(sAlarmDpe, alertValue);	
    } else {
        dyn_string dsExInfo;
        fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr("unSystemIntegrity", "WRONG_ALARM_DPE") + sAlarmDpe + " value = " + alertValue,"");
						
        if(isFunctionDefined("unMessageText_sendException")) {
   		    unMessageText_sendException("*", "*", "SystemIntegrity", "user", "*", dsExInfo);
        }
	   }     
}
